"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import contextlib
import warnings
from dataclasses import dataclass, field
from typing import Optional, Union

import numpy as np
from cardinality import count

from flex.data.lazy_indexable import LazyIndexable


@dataclass(frozen=True)
class Dataset:
    """Class used to represent the dataset from a node in a Federated Learning enviroment.

    Attributes
    ----------
    X_data: LazyIndexable
        A numpy.array containing the data for the node.
    y_data: LazyIndexable
        A numpy.array containing the labels for the training data. Can be None if working
        on an unsupervised learning task. Default None.
    """

    X_data: LazyIndexable = field(init=True)
    y_data: Optional[LazyIndexable] = field(default=None, init=True)

    def __len__(self):
        return len(self.X_data)

    def __getitem__(self, index):
        if isinstance(index, (int, np.integer)):
            return (
                self.X_data[index],
                self.y_data[index] if self.y_data is not None else None,
            )
        elif isinstance(index, (slice, list)):
            return Dataset(
                self.X_data[index],
                self.y_data[index] if self.y_data is not None else None,
            )
        else:
            raise IndexError(
                f"Indexing with element {index} of type {type(index)} is not supported"
            )

    def __iter__(self):
        return zip(
            self.X_data,
            self.y_data if self.y_data is not None else [None] * len(self),
        )

    def to_torchvision_dataset(self, **kwargs):
        """This function transforms a Dataset into a Torchvision dataset object

        Returns:
        --------
            torvhcision.datasets.VisionDataset: a torchvision dataset with the contents of datasets. \
                Note that transforms should be pased as arguments.
        """
        from .dataset_pt_utils import DefaultVision

        return DefaultVision(self, **kwargs)

    def to_tf_dataset(self):
        """This function is an utility to transform a Dataset object to a tensorflow.data.Dataset object

        Returns:
        --------
            tensorflow.data.Dataset: tf dataset object instanciated using the contents of a Dataset
        """
        from tensorflow import type_spec_from_value
        from tensorflow.data import Dataset as tf_Dataset

        return tf_Dataset.from_generator(
            self.__iter__,
            output_signature=(
                type_spec_from_value(self[0][0]),
                type_spec_from_value(self[0][1]),
            ),
        )

    def to_numpy(self, x_dtype=None, y_dtype=None):
        """Function to return the FlexDataObject as numpy arrays."""

        if self.y_data is None:
            return self.X_data.to_numpy(dtype=x_dtype)
        else:
            return self.X_data.to_numpy(x_dtype), self.y_data.to_numpy(dtype=y_dtype)

    def to_list(self):
        """Function to return the FlexDataObject as list."""
        if self.y_data is None:
            return self.X_data.tolist()
        else:
            return self.X_data.tolist(), self.y_data.tolist()

    @classmethod
    def from_torchvision_dataset(cls, pytorch_dataset):
        """Function to convert an object from torchvision.datasets.* to a FlexDataObject.

        Args:
        -----
            pytorch_dataset (torchvision.datasets.*): a torchvision dataset.

        Returns:
        --------
            Dataset: a FlexDataObject which encapsulates the dataset.
        """

        from flex.data.dataset_pt_utils import FeatureDataset, LabelDataset
        from flex.data.pluggable_datasets import PluggableTorchvision

        if pytorch_dataset.__class__.__name__ not in PluggableTorchvision:
            warnings.warn(
                "The input dataset and arguments are not explicitly supported, therefore they might not work as expected.",
                RuntimeWarning,
            )

        length = count(pytorch_dataset)

        X_data = LazyIndexable(FeatureDataset(pytorch_dataset), length=length)
        y_data = LazyIndexable(LabelDataset(pytorch_dataset), length=length)

        return cls(X_data=X_data, y_data=y_data)

    @classmethod
    def from_tfds_image_dataset(cls, tfds_dataset):
        """Function to convert a dataset from tensorflow_datasets to a FlexDataObject.

        Args:
        -----
            tdfs_dataset (tf.data.Datasets): a tf dataset

        Returns:
        --------
            Dataset: a FlexDataObject which encapsulates the dataset.
        """

        if not isinstance(tfds_dataset, tuple):
            # unbatch if required
            with contextlib.suppress(ValueError):
                tfds_dataset = tfds_dataset.unbatch()

            # After unbatching, we can't get the length, so we have to get it.
            # To get the length, we use count.
            length = count(tfds_dataset.as_numpy_iterator())
            X_data = LazyIndexable(
                (x for x, _ in tfds_dataset.as_numpy_iterator()), length=length
            )
            y_data = LazyIndexable(
                (y for _, y in tfds_dataset.as_numpy_iterator()), length=length
            )
        else:
            X_data = LazyIndexable(iter(tfds_dataset[0]), length=len(tfds_dataset[0]))
            y_data = LazyIndexable(iter(tfds_dataset[1]), length=len(tfds_dataset[1]))

        return cls(X_data=X_data, y_data=y_data)

    @classmethod
    def from_tfds_text_dataset(
        cls, tfds_dataset, X_columns: list = None, label_columns: list = None
    ):
        """Function to convert a dataset from tensorflow_datasets to a FlexDataObject.

        Args:
        -----
            tdfs_dataset (tf.data.Datasets): a tf dataset loaded.
            X_columns (list): List containing the features (input) of the model.
            label_columns (list): List containing the targets of the model.

        Returns:
        --------
            Dataset: a FlexDataObject which encapsulates the dataset.
        """
        from tensorflow.python.data.ops.dataset_ops import PrefetchDataset

        if isinstance(tfds_dataset, PrefetchDataset):
            # First case: Users used load func with batch_size != -1 or without indicating the batch_size
            length = len(tfds_dataset)
            if not isinstance(tfds_dataset, tuple):
                with contextlib.suppress(ValueError):
                    tfds_dataset.unbatch()
            if X_columns is None:
                X_data_generator = iter(tfds_dataset.as_numpy_iterator())
            elif len(X_columns) == 1:
                X_data_generator = (
                    tuple(map(row.get, X_columns))[0]
                    for row in tfds_dataset.as_numpy_iterator()
                )
            else:
                X_data_generator = (
                    tuple(map(row.get, X_columns))
                    for row in tfds_dataset.as_numpy_iterator()
                )
            X_data = LazyIndexable(X_data_generator, length=length)

            if label_columns is None:
                y_data = None
            elif len(label_columns) == 1:
                y_data_generator = (
                    tuple(map(row.get, label_columns))[0]
                    for row in tfds_dataset.as_numpy_iterator()
                )
                y_data = LazyIndexable(y_data_generator, length=length)
            else:
                y_data_generator = (
                    tuple(map(row.get, label_columns))
                    for row in tfds_dataset.as_numpy_iterator()
                )
                y_data = LazyIndexable(y_data_generator, length=length)
        else:  # User used batch_size=-1 when using the load function
            if X_columns is None:
                X_data_generator = iter(map(tfds_dataset.get, tfds_dataset.keys()))
            else:
                X_data_generator = iter(map(tfds_dataset.get, X_columns))
            X_data = LazyIndexable(X_data_generator, length=len(tfds_dataset))

            if label_columns is None:
                y_data = None
            else:
                y_data_generator = iter(map(tfds_dataset.get, label_columns))
                y_data = LazyIndexable(y_data_generator, length=len(tfds_dataset))

        return cls(X_data=X_data, y_data=y_data)

    @classmethod
    def from_huggingface_dataset(
        cls,
        hf_dataset,
        X_columns: list = None,
        label_columns: list = None,
    ):
        """Function to conver an arrow dataset from the Datasets package (HuggingFace datasets library)
        to a FlexDataObject.

        Args:
        -----
            hf_dataset (Union[datasets.arrow_dataset.Dataset, str]): a dataset from the dataset library.
            If a string is recieved, it will load the dataset from the HuggingFace repository. When a
            string is given, the split has to be specified in the str variable as follows:
            'dataset;split'. Also, if the string contains a subset, for those datasets that have
            multiple subsets for differents tasks, it may be given as follow: 'dataset;subset;split',
            so we can download the dataset and the desired subset and split.
            X_columns (list): List containing the features names for training the model
            label_columns (list): List containing the name or names of the label column

        Returns:
        --------
            Dataset: a FlexDataObject which encapsulates the dataset.
        """
        from flex.data.pluggable_datasets import PluggableHuggingFace

        try:
            name_checker = ""
            if isinstance(hf_dataset, str):
                from datasets import load_dataset

                hf_dataset = hf_dataset.split(";")
                if len(hf_dataset) == 2:
                    name, split = hf_dataset
                    subset = None
                elif len(hf_dataset) == 3:
                    name, subset, split = hf_dataset
                try:
                    hf_dataset = (
                        load_dataset(name, split=split)
                        if subset is None
                        else load_dataset(
                            name, subset, split=split, ignore_verifications=True
                        )
                    )
                except Exception as err:
                    print(
                        f"Couldn't download the dataset from the HuggingFace datasets: {err}"
                    )

                name_checker = (
                    f"{name.upper()}_{subset.upper()}_HF"
                    if subset is not None
                    else f"{name.upper()}_HF"
                )
            else:
                name_checker = hf_dataset.info.builder_name
            if name_checker not in PluggableHuggingFace.__members__.keys():
                warnings.warn(
                    "The input dataset and arguments are not explicitly supported, therefore they might not work as expected.",
                    RuntimeWarning,
                )

        except Exception:
            warnings.warn(
                "The input dataset doesn't have the property dataset.info.builder_name or the str format is not correct, so we can't check if is supported or not. Therefore, it might not work as expected.",
                RuntimeWarning,
            )
        length = count(hf_dataset)

        if X_columns is None:
            X_data_generator = iter(
                zip(*map(hf_dataset.__getitem__, hf_dataset.features))
            )
        elif len(X_columns) == 1:
            X_data_generator = (
                i for x in map(hf_dataset.__getitem__, X_columns) for i in x
            )
        else:
            X_data_generator = iter(zip(*map(hf_dataset.__getitem__, X_columns)))

        X_data = LazyIndexable(X_data_generator, length=length)

        if label_columns is None:
            y_data = None
        elif len(label_columns) == 1:
            y_data_generator = (
                i for x in map(hf_dataset.__getitem__, label_columns) for i in x
            )
            y_data = LazyIndexable(y_data_generator, length=length)
        else:
            y_data_generator = iter(zip(*map(hf_dataset.__getitem__, label_columns)))
            y_data = LazyIndexable(y_data_generator, length=length)
        return cls(X_data=X_data, y_data=y_data)

    @classmethod
    def from_torchtext_dataset(cls, pytorch_text_dataset):
        """Function to convert an object from torchtext.datasets.* to a FlexDataObject.
            It is mandatory that the dataset contains at least the following transform:
            torchtext.transforms.ToTensor()

        Args:
        -----
            pytorch_text_dataset (torchtext.datasets.*): a torchtext dataset

        Returns:
        --------
            Dataset: a FlexDataObject which encapsulates the dataset.
        """

        from flex.data.pluggable_datasets import PluggableTorchtext

        if pytorch_text_dataset.__class__.__name__ not in PluggableTorchtext:
            warnings.warn(
                "The input dataset and arguments are not explicitly supported, therefore they might not work as expected.",
                RuntimeWarning,
            )
        try:
            length = len(pytorch_text_dataset)
        except TypeError:
            y_data = [label for label, _ in pytorch_text_dataset]
            length = len(y_data)
        X_data = LazyIndexable(
            (text for _, text in pytorch_text_dataset), length=length
        )
        y_data = LazyIndexable(y_data, length=length)

        return cls(X_data=X_data, y_data=y_data)

    @classmethod
    def from_array(
        cls, X_array: Union[list, np.ndarray], y_array: Union[list, np.ndarray] = None
    ):
        """Function that create a Dataset from array-like objects, list and numpy.

        Args:
        -----
            X_array (Union[list, np.ndarray]): Array-like containing X_data.
            y_array (Optional[Union[list, np.ndarray]]): Array-like containing the y_data. Default None.

        Returns:
        --------
            Dataset: a Dataset which encasulates X_array and/or y_array.
        """
        if y_array is not None:
            if not isinstance(X_array, (list, np.ndarray)) or not isinstance(
                y_array, (list, np.ndarray)
            ):
                warnings.warn(  # noqa: B028
                    "X_array or y_array are not a list nor a numpy array. The method might not work as expected.",
                    RuntimeWarning,
                )
        else:
            if not isinstance(X_array, (list, np.ndarray)):
                warnings.warn(  # noqa: B028
                    "X_array is not a list nor a numpy array. The method might not work as expected.",
                    RuntimeWarning,
                )

        X_data = LazyIndexable(X_array, length=len(X_array))
        y_data = (
            None if y_array is None else LazyIndexable(y_array, length=len(y_array))
        )

        return cls(X_data=X_data, y_data=y_data)

    def validate(self):
        """Function that checks whether the object is correct or not."""
        try:
            y_data_length = len(self.y_data)
        except TypeError:
            y_data_length = self.y_data.shape[0]
        if self.y_data is not None and len(self) != y_data_length:
            raise ValueError(
                f"X_data and y_data must have equal lenght. X_data has {len(self)} elements and y_data has {y_data_length} elements."
            )



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from flex.data.dataset import Dataset
from flex.data.fed_dataset import FedDataset
from flex.data.fed_dataset_config import FedDatasetConfig
from flex.data.fed_data_distribution import FedDataDistribution
from flex.data.preprocessing_utils import normalize
from flex.data.preprocessing_utils import one_hot_encoding
from flex.data.pluggable_datasets import PluggableTorchtext
from flex.data.pluggable_datasets import PluggableTorchvision
from flex.data.pluggable_datasets import PluggableHuggingFace
from flex.data.lazy_indexable import LazyIndexable


"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import pytest

from flex.actors.actors import FlexActors
from flex.actors.role import FlexRole


@pytest.fixture(name="actors_cs")
def fixture_actors():
    client1_id = "client1"
    role1 = FlexRole.client
    client2_id = "client2"
    role2 = FlexRole.client
    client3_id = "client3"
    role3 = FlexRole.client
    aggregator_id = "aggregator"
    role4 = FlexRole.aggregator
    server_id = "server"
    role5 = FlexRole.server
    return FlexActors(
        {
            client1_id: role1,
            client2_id: role2,
            client3_id: role3,
            aggregator_id: role4,
            server_id: role5,
        }
    )


@pytest.fixture(name="p2p")
def fixture_peer_to_peer():
    actor1_id = "actor1"
    role1 = FlexRole.server_aggregator_client
    actor2_id = "actor2"
    role2 = FlexRole.server_aggregator_client
    actor3_id = "actor3"
    role3 = FlexRole.server_aggregator_client
    actor4_id = "actor4"
    role4 = FlexRole.server_aggregator_client
    actor5_id = "actor5"
    role5 = FlexRole.server_aggregator_client
    return FlexActors(
        {
            actor1_id: role1,
            actor2_id: role2,
            actor3_id: role3,
            actor4_id: role4,
            actor5_id: role5,
        }
    )


class TestActors(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_simple_actors_client_server(self, actors_cs):
        self._actors_cs = actors_cs

    @pytest.fixture(autouse=True)
    def _fixture_peer_to_peer(self, p2p):
        self._p2p = p2p

    def test_check_compatibility_method(self):
        # Client-Server Architecture
        assert self._actors_cs.check_compatibility("client1", "client2") is False
        assert self._actors_cs.check_compatibility("client1", "server") is False
        assert self._actors_cs.check_compatibility("client1", "aggregator") is True
        assert self._actors_cs.check_compatibility("aggregator", "client3") is False
        assert self._actors_cs.check_compatibility("aggregator", "server") is True
        assert self._actors_cs.check_compatibility("server", "aggregator") is False
        # Peer-To-Peer Architecture
        assert self._p2p.check_compatibility("actor1", "actor2") is True
        assert self._p2p.check_compatibility("actor1", "actor3") is True
        assert self._p2p.check_compatibility("actor1", "actor4") is True
        assert self._p2p.check_compatibility("actor1", "actor5") is True
        assert self._p2p.check_compatibility("actor2", "actor3") is True
        assert self._p2p.check_compatibility("actor2", "actor4") is True
        assert self._p2p.check_compatibility("actor2", "actor5") is True
        assert self._p2p.check_compatibility("actor3", "actor4") is True
        assert self._p2p.check_compatibility("actor3", "actor5") is True
        assert self._p2p.check_compatibility("actor4", "actor5") is True
        assert self._p2p.check_compatibility("actor5", "actor3") is True

    def test_set_function(self):
        client4_id = "client4"
        role_client4 = FlexRole.client
        self._actors_cs[client4_id] = role_client4
        assert client4_id in self._actors_cs.keys()
        assert self._actors_cs[client4_id] == role_client4



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from collections import UserDict
from copy import deepcopy
from functools import partial
from typing import Any, Callable, Hashable, List, Optional

from multiprocess import Pool

from flex.common.utils import check_min_arguments
from flex.data.dataset import Dataset
from flex.data.preprocessing_utils import normalize, one_hot_encoding


class FedDataset(UserDict):
    """Class that represents a federated dataset for the Flex library.
    The dataset contains the ids of the nodes and the dataset associated
    with each node.

    Attributes
    ----------
    data (collections.UserDict): The structure is a dictionary
        with the node ids as keys and the dataset as value.
    """

    def __setitem__(self, key: Hashable, item: Dataset) -> None:
        self.data[key] = item

    def get(self, key: Hashable, default: Optional[Any] = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default

    def apply(
        self,
        func: Callable,
        node_ids: List[Hashable] = None,
        num_proc: int = 1,
        **kwargs,
    ):
        r"""This function lets apply a custom function to the FlexDataset in parallel.

        The \**kwargs provided to this function are all the kwargs of the custom function provided by the node.

        Args:
        -----
            func (Callable, optional): Function to apply to preprocess the data.
            node_ids (List[Hashtable], optional): List containig the the node ids where func will
            be applied. Each element of the list must be hashable and part of the FlexDataset. Defaults to None.
            num_proc (int, optional): Number of processes to parallelize, negative values are ignored. Default to 1

        Returns:
        --------
            FedDataset: The modified FlexDataset.

        Raises:
        -------
            ValueError: All node ids given must be in the FlexDataset.

        """

        if node_ids is None:
            node_ids = list(self.keys())
        elif isinstance(node_ids, str):
            if node_ids not in self.keys():
                raise ValueError("All node ids given must be in the FedDataset.")
        elif any(node not in self.keys() for node in node_ids):
            raise ValueError("All nodes ids given must be in the FedDataset.")

        error_msg = f"The provided function: {func.__name__} is expected to have at least 1 argument/s."
        assert check_min_arguments(func, min_args=1), error_msg

        # if any(self[i].X_data._is_generator for i in node_ids):
        #     raise NotImplementedError("LazyIndexable with generators will be supported soon")

        if num_proc < 2:
            updates = self._map_single(func, node_ids, **kwargs)
        else:
            f = partial(self._map_single, func)
            updates = self._map_parallel(f, node_ids, num_proc=num_proc, **kwargs)

        new_fld = deepcopy(
            self
        )  # seguramente solo haga falta copiar los que no están en node_ids
        new_fld.update(updates)
        return new_fld

    def _map_parallel(
        self,
        func: Callable,
        node_ids: List[Hashable],
        num_proc: int = 2,
        **kwargs,
    ):
        """This function lets apply a custom function to the FlexDataset in parallel.

        The  **kwargs provided to this function are the kwargs of the custom function provided by the node.

        Args:
        -----
            fld (FedDataset): FlexDataset containing all the data from the nodes.
            func (Callable): Function to apply to preprocess the data.
            node_ids (List[Hashtable]): List containig the the nodes id where func will
            be applied. Each element of the list must be hashable and part of the FlexDataset
            num_proc (int): Number of processes to parallelize, negative values are ignored. Default to 2

        Returns:
        --------
            FedDataset: The modified FlexDataset.

        """

        updates = {}
        f = partial(func, **kwargs)  # bind **kwargs arguments to each call

        with Pool(processes=num_proc) as p:
            for i in p.imap(f, node_ids):
                updates.update(i)

        return updates

    def _map_single(
        self,
        func: Callable,
        node_ids: List[Hashable],
        **kwargs,
    ):
        """This function lets apply a custom function to the FlexDataset secuentially.

        This functions will be used by default in the map function, because of the error
        generated by a bug with the multiprocessing library. If you want to check the error
        to try to use the _map_parallel

        The *args and the **kwargs provided to this function are all the args and kwargs
        of the custom function provided by the node.

        Args:
        -----
            func (Callable): Function to apply to preprocess the data.
            node_ids (List[Hashtable]): List containig the the node ids where func will
            be applied. Each element of the list must be hashable and part of the FlexDataset.

        Returns:
        --------
            FedDataset: The modified FlexDataset.
        """
        if not isinstance(node_ids, list):
            node_ids = [node_ids]

        return {node_id: func(self[node_id], **kwargs) for node_id in node_ids}

    def normalize(
        self,
        node_ids: List[Hashable] = None,
        num_proc: int = 0,
        *args,
        **kwargs,
    ):
        """Function that normalize the data over the nodes.

        Args:
        -----
            fld (FedDataset): FlexDataset containing all the data from the nodes.
            node_ids (List[Hashtable], optional): List containig the nodes id whether
            to normalize the data or not. Each element of the list must be hashable. Defaults to None.
            num_proc (int, optional): Number of processes to paralelize. Default to None (Use all).

        Returns:
        --------
            FedDataset: The FlexDataset normalized.
        """
        return self.apply(normalize, node_ids, num_proc, *args, **kwargs)

    def one_hot_encoding(
        self,
        node_ids: List[Hashable] = None,
        num_proc: int = 0,
        *args,
        **kwargs,
    ):
        """Function that apply one hot encoding to the node labels.

        Args:
        -----
            fld (FedDataset): FlexDataset containing all the data from the nodes.
            node_ids (List[Hashtable], optional): List containing the nodes id whether
            to normalize the data or not. Each element of the list must be hashable. Defaults to None.
            num_proc (int, optional): Number of processes to paralelize. Default to None (Use all).

        Returns:
        --------
            FedDataset: The FlexDataset normalized.
        """
        return self.apply(one_hot_encoding, node_ids, num_proc, *args, **kwargs)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from setuptools import find_packages, setup


TF_requires = ["tensorflow<2.11", # https://github.com/tensorflow/tensorflow/issues/58973
                "tensorflow_datasets", 
                "tensorflow_hub"
        ]
PT_requires = ["torch", 
                "torchvision", 
                "torchtext", 
                "torchdata",
                "portalocker",
        ]
HF_requires = ["datasets"]

DIST_requires = ["grpcio", "protobuf"]

setup(
        name="flexible-fl",
        version="0.6.2",
        author="Jimenez-Lopez Daniel, Argente-Garrido Alberto",
        author_email="xehartnort@gmail.com, albertoargentedcgarrido@gmail.com",
        long_description=open("README.md", "r", encoding="utf-8").read(),
        long_description_content_type="text/markdown",
        keywords="FL federated-learning flexible",
        url="https://github.com/FLEXible-FL/FLEX-framework",
        packages=find_packages(),
        install_requires=["numpy",
                        "multiprocess",
                        "scikit-learn",
                        "cardinality",
                        "sultan",
                        "tqdm",
                        "scipy",
                        "gdown",
                        "tensorly"
                        ],
        extras_require={
                "tensorflow": TF_requires,
                "pytorch": PT_requires,
                "hugginface": HF_requires,
                "distributed": DIST_requires,
                "develop": ["pytest",
                        "pytest-asyncio",
                        "pytest-cov",
                        "pytest-xdist",
                        "coverage",
                        "jinja2",
                        *TF_requires,
                        *PT_requires,
                        *HF_requires,
                        *DIST_requires,
                        "grpcio-tools",
                        ],
        },
        python_requires=">=3.8.10",
        classifiers=[
                "Development Status :: 4 - Beta",
                "Intended Audience :: Science/Research",
                "Topic :: Software Development :: Build Tools",
                # Pick your license as you wish
                "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
                # Specify the Python versions you support here. In particular, ensure
                # that you indicate you support Python 3. These classifiers are *not*
                # checked by 'pip install'. See instead 'python_requires' below.
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: 3.8",
                "Programming Language :: Python :: 3.9",
                "Programming Language :: Python :: 3.10",
                "Programming Language :: Python :: 3.11",
                "Programming Language :: Python :: 3 :: Only",
        ],
)



import threading
import unittest
from time import sleep

import numpy as np

from flex.data.dataset import Dataset
from flex.distributed import ClientBuilder, Server
from flex.model.model import FlexModel
from flex.pool.decorators import collect_clients_weights, set_aggregated_weights

addr, port = "localhost", 8080


@collect_clients_weights
def collect_weights(model):
    return [model["model"]["weights1"], model["model"]["weights2"]]


@set_aggregated_weights
def set_weights(model, weights):
    model["model"]["weights1"] = weights[0]
    model["model"]["weights2"] = weights[1]


def train(model, data):
    model["model"]["weights2"] = [7, 8, 9]
    return {"loss": 0.1}


def eval(model, data):
    return {"accuracy": 0.5}


class TestServer(unittest.TestCase):
    @staticmethod
    def run_client():
        model = FlexModel()
        model["model"] = {"weights1": [1, 2, 3], "weights2": [4, 5, 6]}

        dataset = Dataset.from_array([[1, 2], [3, 4]], [0, 1])

        client = (
            ClientBuilder()
            .collect_weights(collect_weights)
            .model(model)
            .set_weights(set_weights)
            .train(train)
            .eval(eval, dataset)
            .dataset(dataset)
            .build()
        )
        client.run(f"{addr}:{port}")

    def test_server_run_called_twice(self):
        server = Server()
        server.run(addr, port)
        with self.assertRaises(RuntimeError):
            server.run(addr, port)

    def test_server_integration_test(self):
        server = Server()
        server.run(addr, port)

        client = threading.Thread(target=self.run_client)
        client.start()
        sleep(1)

        ids = server.get_ids()
        self.assertEqual(ids, ["0"], "Client ID is not correct")
        # Collect weights from the client
        server.ping()
        weights = server.collect_weights(node_ids=ids)
        self.assertTrue(
            all(
                [
                    np.array_equal(w, x)
                    for w, x in zip(
                        weights[0], [np.array([1, 2, 3]), np.array([4, 5, 6])]
                    )
                ]
            )
        )

        # set weights
        new_weights = [np.array([10, 11, 12]), np.array([13, 14, 15])]
        server.send_weights(weights=new_weights, node_ids=ids)
        weights = server.collect_weights(node_ids=ids)
        self.assertTrue(
            all([np.array_equal(w, x) for w, x in zip(weights[0], new_weights)])
        )

        # train
        metrics = server.train(node_ids=ids)
        self.assertIsInstance(metrics, dict)
        self.assertTrue("0" in metrics)
        self.assertTrue("loss" in metrics["0"])
        self.assertAlmostEqual(metrics["0"]["loss"], 0.1)
        # Check that train has changed weights
        weights = server.collect_weights(node_ids=ids)
        self.assertTrue(
            all(
                [
                    np.array_equal(w, x)
                    for w, x in zip(
                        weights[0], [np.array([10, 11, 12]), np.array([7, 8, 9])]
                    )
                ]
            ),
            f"weights: {weights}, expected: {[np.array([10, 11, 12]), np.array([7, 8, 9])]}",
        )

        # Eval
        metrics = server.eval(node_ids=ids)
        self.assertIsInstance(metrics, dict)
        self.assertTrue("0" in metrics)
        self.assertTrue("accuracy" in metrics["0"])
        self.assertAlmostEqual(metrics["0"]["accuracy"], 0.5)

        server.stop()
        client.join()



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from enum import Enum, EnumMeta


class PluggableDataset(EnumMeta):
    def __contains__(self, item):
        return item in list(self.members())


class PluggableDatasetString(EnumMeta):
    def __contains__(self, item):
        return item in [k.value[0] for k in self.__members__.values()]


class PluggableTorchtext(Enum, metaclass=PluggableDataset):
    """Class containing all the pluggable datasets to a Dataset without any preprocessing needed.

    Any other dataset from the TorchText library will need further preprocessing.

    Args:
    -----
        Enum (enum): torchtext class for each dataset than can be accepted on our platform.
    """

    def members():
        from torchtext import datasets

        yield datasets.AG_NEWS.__name__
        yield datasets.AmazonReviewFull.__name__
        yield datasets.AmazonReviewPolarity.__name__
        yield datasets.DBpedia.__name__
        yield datasets.YahooAnswers.__name__
        yield datasets.YelpReviewFull.__name__
        yield datasets.YelpReviewPolarity.__name__


class PluggableTorchvision(Enum, metaclass=PluggableDataset):

    """Class containing all the pluggable datasets to a Dataset without any preprocessing needed.

    Any other dataset from the Torchvision library will need further preprocessing.

    Args:
    -----
        Enum (enum): torchvision class for each dataset than can be accepted on our platform.
    """

    def members():
        from torchvision import datasets

        yield datasets.WIDERFace.__name__
        yield datasets.Food101.__name__
        yield datasets.CelebA.__name__
        yield datasets.CLEVRClassification.__name__
        yield datasets.Country211.__name__
        yield datasets.FGVCAircraft.__name__
        yield datasets.GTSRB.__name__
        yield datasets.Kitti.__name__
        yield datasets.Flowers102.__name__
        yield datasets.StanfordCars.__name__
        yield datasets.LFWPeople.__name__
        yield datasets.Caltech256.__name__
        yield datasets.EuroSAT.__name__
        yield datasets.CIFAR10.__name__
        yield datasets.CIFAR100.__name__
        yield datasets.MNIST.__name__
        yield datasets.SUN397.__name__
        yield datasets.SEMEION.__name__
        yield datasets.Omniglot.__name__
        yield datasets.KMNIST.__name__
        yield datasets.FashionMNIST.__name__
        yield datasets.OxfordIIITPet.__name__
        yield datasets.STL10.__name__
        yield datasets.PCAM.__name__
        yield datasets.Caltech101.__name__
        yield datasets.QMNIST.__name__
        yield datasets.SVHN.__name__
        yield datasets.DTD.__name__
        yield datasets.USPS.__name__
        yield datasets.RenderedSST2.__name__
        yield datasets.INaturalist.__name__
        yield datasets.EMNIST.__name__


class PluggableHuggingFace(Enum, metaclass=PluggableDatasetString):
    """Class containing some datasets that can be loaded to FLEXible. Other datasets
    can be plugged in, but it requires a special configuration, i.e., glue-cola. This
    is more about the user using correctly the arguments on the load_dataset function
    from huggingface datasets than a problem of our platform, so the user can easy-use
    other datasets.

    We show some example datasets that can be loaded using the function
    FedDataDistribution.from_config_with_huggingface_dataset just giving a config
    and the string associated to each dataset from the Enum defined.

    We selected this dataset as we can automatice the process of loading this datasets,
    but our framework support almost all the datasets, as they can be loaded as numpy
    arrays. We only show supports to this datasets as we can load the dataset
    as follows: dataset = load_dataset(name, split='train').

    There are some datasets that need extra parameters like the version of the dataset,
    or that don't have any split. This must be used by the user previously to load
    the dataset into FLEXible, but it will be easy and fast, as the user just
    need to select the X_train-y_train as np.arrays.

    Args:
    -----
        Enum (enum): Tuple containing name, X_columns and y_columns to use in the
        load_dataset function.
    """

    IMDB_HF = ("imdb", "text", "label")
    AG_NEWS_HF = ("ag_news", "text", "label")
    TWEET_EVAL_EMOJI_HF = ("tweet_eval", "text", "label")  # We support all subsets
    ROTTEN_TOMATOES_HF = ("rotten_tomatoes", "text", "label")
    GLUE_COLA_HF = ("glue", "sentence", "label")
    FINANCIAL_PHRASEBANK_HF = (
        "financial_phrasebank",
        "sentence",
        "label",
    )  # We support all subsets"
    SQUAD_HF = ("squad", ["context", "question"], "answers")
    APPREVIEWS_HF = ("app_reviews", "review", "star")
    AMAZON_POLARITY_HF = ("amazon_polarity", ["title", "content"], "label")


# class PluggableDatasetsTensorFlowText(Enum, metaclass=PluggableDataset):
#     """Class containing some datasets that can be loaded to FLEXible. Other datasets
#     can be plugged in, but it requires a special configuration, i.e., glue-cola. This
#     is more about the user using correctly the arguments on the load_dataset function
#     from huggingface datasets than a problem of our platform, so the user can easy-use
#     other datasets.

#     Args:
#         Enum (enum): Tuple containing name, X_columns and y_columns to use in the
#         load_dataset function.
#     """

#     AG_NEWS_TF = ("ag_news_subset", ["title", "description"], ["label"])
#     GLUE_TF = ("glue", ["sentence"], ["label"])
#     ASSET_TF = ("asset", ["original"], ["simplifications"])
#     SQUAD_TF = ("squad", ["title", "question", "context"], ["answers"])  # Wonk work
#     COQA_TF = ("coqa", ["questions", "source", "story"], ["answers"])  # Won't work



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import pytest

from flex.data import FedDataDistribution, FedDatasetConfig


@pytest.fixture(name="basic_config")
def fixture_simple_fex_data_object():
    return FedDatasetConfig(
        seed=0,
        n_nodes=2,
        replacement=False,
        node_ids=["client_0", "client_1"],
    )


class TestFlexDataDistribution(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_simple_fex_data_object(self, basic_config):
        self._config = basic_config

    def test_from_pytorch_text_dataset(self):
        from torchtext.datasets import AG_NEWS

        data = AG_NEWS(split="train")
        flex_dataset = FedDataDistribution.from_config_with_torchtext_dataset(
            data, self._config
        )
        assert len(flex_dataset) == self._config.n_nodes
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])

    def test_from_huggingface_text_dataset(self):
        from datasets import load_dataset

        data = load_dataset("ag_news", split="train")
        X_columns = "text"
        label_column = "label"
        flex_dataset = FedDataDistribution.from_config_with_huggingface_dataset(
            data, self._config, X_columns, label_column, lazy=False
        )
        assert len(flex_dataset) == self._config.n_nodes
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])
        assert (
            len(flex_dataset["client_0"]) + len(flex_dataset["client_1"])
            == data.num_rows
        )
